""" DIRAC Benchmark 2012 by Ricardo Graciani, and wrapper functions to
    run multiple copies in parallel by Andrew McNab.
    This file (dirac_benchmark.py) is intended to be the ultimate upstream
    shared by different users of the DIRAC Benchmark 2012 (DB12). The
    canonical version can be found at https://github.com/DIRACGrid/DB12
    This script can either be imported or run from the command line:
    ./dirac_benchmark.py NUMBER
    where NUMBER gives the number of benchmark processes to run in parallel.
    Run  ./dirac_benchmark.py help  to see more options.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import argparse
import json
from pkg_resources import get_distribution, DistributionNotFound

from db12 import single_dirac_benchmark
from db12 import multiple_dirac_benchmark
from db12 import wholenode_dirac_benchmark

try:
    VERSION = get_distribution("db12").version
except DistributionNotFound:
    pass


def dump_as_json(filename, output):
    """Function to save result to a json file

    :param str filename: path to a json file
    :param dict output: content to dump in the json file
    """
    with open(filename, "w") as outfile:
        json.dump(output, outfile)


def single_dirac_benchmark_cli(args):
    """Function that calls single_dirac_benchmark and prints
    its results/calls dump_as_json"""
    result = single_dirac_benchmark(
        iterations_num=args.iterations, correction=args.correction
    )["NORM"]
    if args.json:
        dump_as_json(args.json, result)
    else:
        print(result)


def multiple_dirac_benchmark_cli(args):
    """Function that calls multiple_dirac_benchmark and prints
    its results/calls dump_as_json"""
    result = multiple_dirac_benchmark(
        copies=args.copies,
        iterations_num=args.iterations,
        extra_iter=args.extra_iteration,
        correction=args.correction,
    )
    if args.json:
        dump_as_json(args.json, result)
    else:
        print(
            result["copies"],
            result["sum"],
            result["arithmetic_mean"],
            result["geometric_mean"],
            result["median"],
        )
        print(" ".join([str(k) for k in result["raw"]]))


def wholenode_dirac_benchmark_cli(args):
    """Function that calls wholenode_dirac_benchmark and prints
    its results/calls dump_as_json"""
    result = wholenode_dirac_benchmark(
        iterations_num=args.iterations,
        extra_iter=args.extra_iteration,
        correction=args.correction,
    )
    if args.json:
        dump_as_json(args.json, result)
    else:
        print(
            result["copies"],
            result["sum"],
            result["arithmetic_mean"],
            result["geometric_mean"],
            result["median"],
        )
        print(" ".join([str(j) for j in result["raw"]]))


def main():
    """Uses the functions within dirac_benchmark.py to run the DB12 benchmark from
    the command line: single, jobslots, multiple and wholenode.
    By default, one benchmarking iteration is run, in addition to the initial
    iteration which DB12 runs and ignores to avoid ramp-up effects at the start.
    The number of benchmarking iterations can be increased using the --iterations
    option. Additional iterations which are also ignored can be added with the
    --extra-iteration option to avoid tail effects. In this case copies which
    finish early run additional iterations until all the measurements finish.
    The copies (ie an integer) argument causes multiple copies of the benchmark to
    be run in parallel. Unless the token "single" is used, the script prints the
    following results to two lines on stdout:
    COPIES SUM ARITHMETIC-MEAN GEOMETRIC-MEAN MEDIAN
    RAW-RESULTS"""

    help_string = """Run the DB12 benchmark from the command line.
Unless the token "single" is used, the script prints the following results to
two lines on stdout:
COPIES SUM ARITHMETIC-MEAN GEOMETRIC-MEAN MEDIAN
RAW-RESULTS
"""

    # Common arguments: --iterations, --json, --no-correction
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "--iterations",
        nargs="?",
        type=int,
        help="number of iterations to perform",
        default=1,
    )
    parent_parser.add_argument("--json", help="generate json files", type=str)
    parent_parser.add_argument(
        "--no-correction",
        help="Disable norm correction",
        dest="correction",
        action="store_false",
    )

    # Main parser
    parser = argparse.ArgumentParser(description=help_string)
    parser.add_argument("--version", action="version", version=VERSION, default="")

    # Subparser: single benchmark
    subparsers = parser.add_subparsers()
    parser_single = subparsers.add_parser(
        "single", help="get normalized power of one CPU", parents=[parent_parser]
    )
    parser_single.set_defaults(func=single_dirac_benchmark_cli)

    # Subparser: wholenode benchmark
    parser_wholenode = subparsers.add_parser(
        "wholenode",
        help="Run as many copies as needed to occupy the whole machine",
        parents=[parent_parser],
    )
    parser_wholenode.add_argument(
        "--extra-iteration",
        help="whether an extra iteration is needed",
        action="store_true",
    )
    parser_wholenode.set_defaults(func=wholenode_dirac_benchmark_cli)

    # Subparser: multiple benchmark
    parser_multiple = subparsers.add_parser(
        "multiple", help="Run multiple copies in parallel", parents=[parent_parser]
    )
    parser_multiple.add_argument("copies", type=int, help="number of copies")
    parser_multiple.add_argument(
        "--extra-iteration",
        help="whether an extra iteration is needed",
        action="store_true",
    )
    parser_multiple.set_defaults(func=multiple_dirac_benchmark_cli)

    # Get params and execute the selected function
    args = parser.parse_args()

    if hasattr(args, "func") and args.func:
        args.func(args)
    else:
        parser.print_help()


#
# If we run as a command
#
if __name__ == "__main__":
    main()



"""Unit test for DB12"""
from __future__ import absolute_import
from __future__ import division
import pytest

from db12 import single_dirac_benchmark
from db12 import multiple_dirac_benchmark
from db12 import wholenode_dirac_benchmark


@pytest.mark.parametrize(
    "copies, iterations, extra_iteration",
    [
        ("single", 1, False),
        ("single", 2, True),
        ("wholenode", 1, False),
        ("wholenode", 2, True),
        (2, 1, False),
        (3, 2, True),
    ],
)
def test_dirac_benchmark(copies, iterations, extra_iteration):
    """Testing each function in the DB12 to make sure that:
    1- It produces the same output when run with the same input
    2- It produces the same output when run with a different number of iterations
    """
    k = 0
    threshold = 20 / 100

    if copies is None or copies == "single":
        result = single_dirac_benchmark(iterations_num=iterations)["NORM"]
        result2 = single_dirac_benchmark(iterations_num=iterations)["NORM"]
        result3 = single_dirac_benchmark(iterations_num=iterations + 1)["NORM"]

        assert abs(result2 - result) <= result * threshold
        assert abs(result3 - result) <= result * threshold

        assert result >= 0
        assert result < 100

    elif copies == "wholenode":
        result = wholenode_dirac_benchmark(
            iterations_num=iterations, extra_iter=extra_iteration
        )
        result2 = wholenode_dirac_benchmark(
            iterations_num=iterations, extra_iter=extra_iteration
        )
        result3 = wholenode_dirac_benchmark(
            iterations_num=iterations + 1, extra_iter=extra_iteration
        )

        assert result["geometric_mean"] >= 0 and result["geometric_mean"] < 100
        assert result["arithmetic_mean"] >= 0 and result["arithmetic_mean"] < 100
        assert result["median"] >= 0 and result["median"] < 100

        for i in result["raw"]:
            assert abs(i - result2["raw"][k]) <= result2["raw"][k] * threshold
            assert i >= 0
            assert i < 100
            k = k + 1

        k = 0
        for i in result2["raw"]:
            assert abs(i - result3["raw"][k]) <= result3["raw"][k] * threshold
            assert i >= 0
            assert i < 100
            k = k + 1
    else:
        result = multiple_dirac_benchmark(
            copies=int(copies), iterations_num=iterations, extra_iter=extra_iteration
        )
        result2 = multiple_dirac_benchmark(
            copies=int(copies), iterations_num=iterations, extra_iter=extra_iteration
        )
        result3 = multiple_dirac_benchmark(
            copies=int(copies),
            iterations_num=iterations + 1,
            extra_iter=extra_iteration,
        )

        assert result["geometric_mean"] >= 0 and result["geometric_mean"] < 100
        assert result["arithmetic_mean"] >= 0 and result["arithmetic_mean"] < 100
        assert result["median"] >= 0 and result["median"] < 100

        for i in result["raw"]:
            assert abs(i - result2["raw"][k]) <= result2["raw"][k] * threshold
            assert i >= 0
            assert i < 100
            k = k + 1

        k = 0
        for i in result2["raw"]:
            assert abs(i - result3["raw"][k]) <= result3["raw"][k] * threshold
            assert i >= 0
            assert i < 100
            k = k + 1



"""Contains all the functions needed to run DB12"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import logging
import json
import sys
import random
import re
import multiprocessing

if sys.version_info[0] < 3:
    # pylint: disable = E, W, R, C
    # python3 range corresponds to xrange in python2
    range = xrange


def single_dirac_benchmark(iterations_num=1, measured_copies=None, correction=True):
    """Get Normalized Power of one CPU in DIRAC Benchmark 2012 units (DB12)

    :param int iterations_num: number of iterations to run
    :param multiprocessing.Value measured_copies: extra iterations to run
    """
    # pylint: disable = too-many-locals
    # This number of iterations corresponds to 1kHS2k.seconds, i.e. 250 HS06 seconds

    iters = int(1000 * 1000 * 12.5)
    calib = 250.0
    m_1 = int(0)
    m_2 = int(0)
    if sys.version_info[0] < 3:
        # pylint: disable = E, W, R, C
        # long type does not exist anymore in python3 but was used in this context with python2
        m_1 = long(0)
        m_2 = long(0)

    p_1 = 0
    p_2 = 0
    # Do one iteration extra to allow CPUs with variable speed (we ignore zeroth iteration)
    # Do one or more extra iterations to avoid tail effects when copies run in parallel
    it_1 = 0
    while (it_1 <= iterations_num) or (
        measured_copies is not None and measured_copies.value > 0
    ):
        if it_1 == 1:
            start = os.times()

        # Now the iterations
        for _j in range(iters):
            t_1 = random.normalvariate(10, 1)
            m_1 += t_1
            m_2 += t_1 * t_1
            p_1 += t_1
            p_2 += t_1 * t_1

        if it_1 == iterations_num:
            end = os.times()
            if measured_copies is not None:
                # Reduce the total of running copies by one
                measured_copies.value -= 1

        it_1 += 1

    cput = sum(end[:4]) - sum(start[:4])
    wall = end[4] - start[4]

    if not cput:
        return None

    norm = calib * iterations_num / cput
    if correction:
        norm = get_norm_correction(norm)

    # Return DIRAC-compatible values
    output = {
        "CPU": cput,
        "WALL": wall,
        "NORM": norm,
        "UNIT": "DB12",
    }
    return output


def single_dirac_benchmark_process(
    result_object, iterations_num=1, measured_copies=None, correction=True
):
    """Run single_dirac_benchmark() in a multiprocessing friendly way

    :param multiprocessing.Value result_object: result to be returned
    :param int iterations_num: number of iterations to run
    :param multiprocessing.Value measured_copies: extra iterations to run
    """

    benchmark_result = single_dirac_benchmark(
        iterations_num=iterations_num,
        measured_copies=measured_copies,
        correction=correction,
    )

    if not benchmark_result or "NORM" not in benchmark_result:
        return

    # This makes it easy to use with multiprocessing.Process
    result_object.value = benchmark_result["NORM"]


def multiple_dirac_benchmark(
    copies=1, iterations_num=1, extra_iter=False, correction=True
):
    """Run multiple copies of the DIRAC Benchmark in parallel

    :param int copies: number of single benchmark to run in parallel
    :param int interations_num: number of iterations to run
    :param bool extra_iter: to know whether it should include extra iterations
    """

    processes = []
    results = []

    if extra_iter:
        # If true, then we run one or more extra iterations in each
        # copy until the number still being meausured is zero.
        measured_copies = multiprocessing.Value("i", copies)
    else:
        measured_copies = None

    # Set up all the subprocesses
    for i in range(copies):
        results.append(multiprocessing.Value("d", 0.0))
        processes.append(
            multiprocessing.Process(
                target=single_dirac_benchmark_process,
                args=(results[i], iterations_num, measured_copies, correction),
            )
        )

    # Start them all off at the same time
    for process in processes:
        process.start()

    # Wait for them all to finish
    for process in processes:
        process.join()

    raw = []
    product = 1.0

    for res in results:
        raw.append(res.value)
        product *= res.value

    raw.sort()

    # Return the list of raw results and various averages
    output = {
        "raw": raw,
        "copies": copies,
        "sum": sum(raw),
        "arithmetic_mean": sum(raw) / copies,
        "geometric_mean": product ** (1.0 / copies),
        "median": raw[(copies - 1) // 2],
    }
    return output


def wholenode_dirac_benchmark(iterations_num=1, extra_iter=False, correction=True):
    """Run as many copies as needed to occupy the whole machine

    :param int interations_num: number of iterations to run
    :param bool extra_iter: to know whether it should include extra iterations
    """

    try:
        copies = multiprocessing.cpu_count()
    except:  # pylint: disable=bare-except
        copies = 1

    return multiple_dirac_benchmark(
        copies=copies,
        iterations_num=iterations_num,
        extra_iter=extra_iter,
        correction=correction,
    )


def get_norm_correction(norm_computed):
    """Apply a factor on the norm depending on the python version used and
    the architecture targeted in order to reproduce the original norm from python2.

    :param float norm_computed: raw norm
    """
    # If python2 is used, then no action is needed
    if sys.version_info[0] < 3:
        return norm_computed

    logging.warn(
        "You are executing DB12 using python3, DB12 score is generally higher than it was with python2"
    )
    logging.warn("Trying to apply a correction...")

    # Get the dictionary of factors
    with open(
        os.path.join(os.path.dirname(__file__), "factors.json"), "r"
    ) as file_object:
        factor_dict = json.load(file_object)

    # Get Python version: if not in the dictionary, no action can be performed
    major, minor, micro = sys.version_info[0:3]
    python_version = "%s.%s.%s" % (major, minor, micro)
    if python_version not in factor_dict.keys():
        logging.warn(
            "Cannot correct the score, return the raw norm: the python version you are using has not been analyzed."
        )
        logging.warn(
            "Please consult https://zenodo.org/record/5647834 for further details"
        )
        return norm_computed

    # Get CPU brand name
    try:
        with open("/proc/cpuinfo", "r") as file_object:
            content = file_object.read()
        cpu_model_name = re.findall("model name\t: ([a-zA-Z]*) ", content)[0]
    except IOError:
        logging.warn(
            "Cannot correct the score, return the raw norm: cannot access CPU information"
        )
        return norm_computed

    factor = factor_dict[python_version].get(cpu_model_name)
    if not factor:
        logging.warn(
            "Cannot correct the score, return the raw norm: the CPU model you are using has not been analyzed."
        )
        logging.warn(
            "Please consult https://zenodo.org/record/5647834 for further details."
        )
        return norm_computed

    logging.info("Applying a factor of %s to the raw norm %s", factor, norm_computed)
    return norm_computed * factor



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="DB12",
    description="DIRAC Benchmark 2012",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/DIRACGrid/DB12",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    project_urls={
        "Bug Tracker": "https://github.com/DIRACGrid/DB12/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2.7",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    package_data={"": ["factors.json"]},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=2.7",
    test_suite="tests",
    entry_points={
        "console_scripts": [
            "db12 = db12.__main__:main",
        ],
    },
)
